import java.util.ArrayList;
import java.util.LinkedList;

public class RPNCalculatorUtil {

    public static final String LEFT_BRACKET = "(";
    public static final String RIGHT_BRACKET = ")";
    private static final String DECIMAL_REGEX = "^([0-9]+\\.?[0-9]*|[0-9]*\\.[0-9]+)$";

    /**
     * Checks if the given token is a decimal or normal number.
     * valid: 1.5232, 2, 3, 0.444
     * @param token the input token
     * @return true if it is a number else false
     */
    public static boolean isNumber(String token) {
        return token.matches(DECIMAL_REGEX);
    }

    /**
     * Parses a infix/postfix expression string into tokens.
     * ASSUMES WHITESPACES between operands, numbers, and brackets
     * Example input: "( 1 + 2 ) / 3 ="
     * output: ["(", "1", "+", "2", ")", "/", "3", "="]
     * @param in parses an infix/postfix expression
     * @return a String array of tokens
     * @throws InvalidInputException if there is an invalid token, which is not supported
     */
    public static String[] parse(String in) throws InvalidInputException {
        String[] tokens = in.split("\\s+");
        if (tokens.length < 1) {
            throw new InvalidInputException("Empty Line");
        }
        if (!tokens[tokens.length - 1].equals("=")) {
            throw new InvalidInputException("No = token at the end of the line");
        }
        for (String token : tokens) {
            if (!(token.matches(DECIMAL_REGEX)
                    || RPNCalculator.isOperator(token)
                    || token.equals(RPNCalculator.END_TOKEN)
                    || token.equals(LEFT_BRACKET)
                    || token.equals(RIGHT_BRACKET))) {
                throw new InvalidInputException(String.format("Token %s has an invalid format", token));
            }
        }
        return tokens;
    }

    /**
     * Transforms an infix expression to a postfix expression.
     * The method implements the Shunting-yard algorithm for the transformation.
     * Example Input: ["(", "1", "+", "2", ")", "/", "3", "="]
     * output: ["1", "2", "+", "3", "/", "="]
     * @param infix the infix expression as a String array of tokens.
     * @return the postfix expression as a String array of tokens.
     * @throws InvalidInputException if a token in the array is not supported or there is a parentheses mismatch.
     */
    public static String[] transformInfixToPostfix(String[] infix) throws RPNException {
        ArrayList<String> output = new ArrayList<>();
        LinkedList<String> operatorStack = new LinkedList<>();

        for (String token : infix) {
            if (isNumber(token)) {
                // The token is a number (e.g. 120).
                //TODO
            } else if (token.equals(LEFT_BRACKET)) {
                // The token is a left bracket.
                //TODO
            } else if (token.equals(RIGHT_BRACKET)) {
                // The token is a right bracket.
                String nextOP = operatorStack.peek();
                while (nextOP != null && !nextOP.equals(LEFT_BRACKET)) {
                    output.add(operatorStack.pop());
                    nextOP = operatorStack.peek();
                }
                if (nextOP == null) {
                    throw new RPNException("Mismatched parentheses");
                }
                operatorStack.pop();
            } else if (RPNCalculator.operatorPrecedence.containsKey(token)) {
                // The token is an operator.
                //TODO
            } else if (token.equals(RPNCalculator.END_TOKEN)) {
                // The token is the end token "=".
                break;
            } else {
                throw new RPNException(String.format("The token %s is not supported", token));
            }
        }
        // TODO return String-array
        return ;
    }

    /**
     * Evaluates a postfix expression with an RPN calculator.
     * @param calculator the RPN calculator to use.
     * @param tokens the postfix expression as an array of tokens.
     * @return the result of the evaluation
     * @throws RPNException if the token is not supported or there was no eval "=" token in the expression.
     */
    public static double evaluatePostfixExpression(RPNCalculator calculator, String[] tokens) throws RPNException {
        for (int i = 0; i < tokens.length; i++) {
			String token = tokens[i];
            if (RPNCalculator.isOperator(token)) {
                //TODO
            } else if (isNumber(token)) {
                //TODO
            } else if (RPNCalculator.resultToken(token)) {
                //TODO
            } else {
                throw new RPNException(String.format("The token %s is not supported", token));
            }
        }
        throw new RPNException("No evaluation token = in the array");
    }

    public static double evaluateInfixExpression(RPNCalculator calculator, String[] tokens) {
        return evaluatePostfixExpression(calculator, transformInfixToPostfix(tokens));
    }
}